package io.gitlab.jfronny.inceptum.cli.commands;

import io.gitlab.jfronny.commons.io.JFiles;
import io.gitlab.jfronny.commons.throwable.ThrowingBiFunction;
import io.gitlab.jfronny.inceptum.cli.*;
import io.gitlab.jfronny.inceptum.common.Utils;
import io.gitlab.jfronny.inceptum.launcher.model.inceptum.InstanceMeta;
import io.gitlab.jfronny.inceptum.launcher.util.ModManager;
import io.gitlab.jfronny.inceptum.launcher.util.ModPath;
import io.gitlab.jfronny.inceptum.launcher.system.mds.IWModDescription;
import io.gitlab.jfronny.inceptum.launcher.system.mds.ModsDirScanner;
import io.gitlab.jfronny.inceptum.launcher.system.source.ModSource;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;

public class ModCommand extends Command {
    //TODO ModAddCommand (search and install)
    public ModCommand() {
        super("Allows managing mods in instances", "", List.of("mod"), List.of(
                new ModListCommand(),
                new ModUpdateCommand(),
                new ModRemoveCommand()
        ));
    }

    @Override
    protected void invoke(CommandArgs args) {
        Utils.LOGGER.error("No known mod command was specified");
    }

    public static class ModListCommand extends BaseInstanceCommand {
        private final boolean filterUpdatable;

        public ModListCommand() {
            this("Lists the mods in an instance", List.of("list", "ls"), List.of(
                    new ModListCommand("List only updatable mods", List.of("updatable"), List.of(), true)
            ), false);
        }

        private ModListCommand(String help, List<String> aliases, List<Command> subCommands, boolean filterUpdatable) {
            super(help, "", aliases, subCommands);
            this.filterUpdatable = filterUpdatable;
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta) throws IOException {
            if (!meta.isFabric()) {
                System.err.println("This is not a fabric instance");
                return;
            }
            System.out.println("Scanning installed mods, this might take a while");
            ModsDirScanner mds = ModsDirScanner.get(instancePath.resolve("mods"), meta);
            mds.runOnce((path, mod) -> {
                boolean hasSources = mod.mod().isPresent() && !mod.mod().get().sources.isEmpty();
                boolean updatable = hasSources && mod.mod().get().sources.values().stream().anyMatch(Optional::isPresent);
                if (filterUpdatable && !updatable) return;
                System.out.println("- " + path.getFileName().toString());
                System.out.println("  " + mod.getName());
                for (String s : mod.getDescription()) {
                    System.out.println("  " + s);
                }
                if (hasSources) {
                    System.out.println("  Sources:");
                    for (Map.Entry<ModSource, Optional<ModSource>> entry : mod.mod().get().sources.entrySet()) {
                        System.out.println("  - " + entry.getKey().getName() + " (" + entry.getKey().getVersion() + ")");
                        System.out.println("    Local: " + entry.getKey().getJarPath().toString());
                        if (entry.getValue().isPresent())
                            System.out.println("    Updatable to: " + entry.getValue().get().getVersion());
                    }
                }
            });
        }
    }

    public static class ModRemoveCommand extends BaseInstanceCommand {
        private final boolean ignoreDependencies;

        public ModRemoveCommand() {
            this("Removes mods from an instance", List.of("remove", "delete", "rm", "del"), List.of(
                    new ModRemoveCommand("Skip dependency checks", List.of("ignore-dependencies"), List.of(), true)
            ), false);
        }

        private ModRemoveCommand(String help, List<String> aliases, List<Command> subCommands, boolean ignoreDependencies) {
            super(help, "<mod file>...", aliases, subCommands);
            this.ignoreDependencies = ignoreDependencies;
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta) throws IOException {
            if (!meta.isFabric()) {
                Utils.LOGGER.error("This is not a fabric instance");
                return;
            }
            if (args.length == 0) {
                Utils.LOGGER.error("You must specify mods to remove");
                return;
            }
            Set<Path> mods = new HashSet<>();
            for (String arg : args) {
                Path p = instancePath.resolve("mods").resolve(arg);
                if (!Files.exists(p)) p = instancePath.resolve("mods").resolve(arg + ".imod");
                if (!Files.exists(p)) {
                    Utils.LOGGER.error("Nonexistant mod file: " + instancePath.resolve("mods").resolve(arg));
                    return;
                }
                mods.add(p);
            }
            ModsDirScanner mds = ModsDirScanner.get(instancePath.resolve("mods"), meta);
            if (!ignoreDependencies && !mds.isComplete()) {
                Utils.LOGGER.error("Scanning mods dir to search for dependencies. This might take a while");
                mds.runOnce((path, mod) -> System.out.println("Scanned " + path));
            }
            for (Path mod : mods) {
                try {
                    ModManager.delete(mds.get(mod), instancePath.resolve("mods"), mds);
                } catch (IOException e) {
                    Utils.LOGGER.error("Could not delete " + mod, e);
                    return;
                }
            }
        }
    }

    public static class ModUpdateCommand extends BaseInstanceCommand {
        private final ThrowingBiFunction<CommandArgs, Path, Set<Path>, IOException> pathSupplier;

        public ModUpdateCommand() {
            this("Update mods in an instance", "<mod file>...", List.of("update", "upgrade", "up"), List.of(
                    new ModUpdateCommand("Update all mods in an instance", "", List.of("all"), List.of(), (args, instancePath) -> Set.copyOf(JFiles.list(instancePath)))
            ), (args, instancePath) -> {
                Set<Path> mods = new HashSet<>();
                for (String arg : args) {
                    Path p = instancePath.resolve("mods").resolve(arg);
                    if (!Files.exists(p)) p = instancePath.resolve("mods").resolve(arg + ".imod");
                    if (!Files.exists(p)) {
                        throw new FileNotFoundException("Nonexistant mod file: " + instancePath.resolve("mods").resolve(arg));
                    }
                    mods.add(p);
                }
                return mods;
            });
        }

        private ModUpdateCommand(String help, String usage, List<String> aliases, List<Command> subCommands, ThrowingBiFunction<CommandArgs, Path, Set<Path>, IOException> pathSupplier) {
            super(help, usage, aliases, subCommands);
            this.pathSupplier = pathSupplier;
        }

        @Override
        protected void invoke(CommandArgs args, Path instancePath, InstanceMeta meta) throws IOException {
            if (!meta.isFabric()) {
                throw new IOException("This is not a fabric instance");
            }
            if (args.length == 0) {
                throw new IllegalArgumentException("You must specify mods to remove");
            }
            Set<Path> mods = pathSupplier.apply(args, instancePath);
            ModsDirScanner mds = ModsDirScanner.get(instancePath.resolve("mods"), meta);
            if (!mds.isComplete()) {
                Utils.LOGGER.error("Scanning mods dir to search for dependencies. This might take a while");
                mds.runOnce((path, mod) -> System.out.println("Scanned " + path));
            }
            for (Path mod : mods) {
                try {
                    Utils.LOGGER.info("Updating " + mod);
                    ModManager.delete(mds.get(mod), instancePath.resolve("mods"), mds);
                    IWModDescription md = mds.get(mod);
                    if (md.mod().isEmpty()) {
                        throw new IOException("Could not load mod description");
                    }
                    boolean found = false;
                    for (Map.Entry<ModSource, Optional<ModSource>> source : md.mod().get().sources.entrySet()) {
                        Optional<ModSource> ms = source.getValue();
                        if (ms.isPresent()) {
                            try {
                                Utils.LOGGER.info("Updating to " + ms.get().getVersion());
                                Path imodPath = md.imod().isPresent() ? md.imod().get() : instancePath.resolve("mods").resolve(ms.get().getShortName() + ModPath.EXT_IMOD);
                                ModManager.DownloadMeta dm = ModManager.download(ms.get(), imodPath, mds);
                                Files.delete(md.path());
                                if (md.imod().isPresent() && Files.exists(md.imod().get()))
                                    Files.delete(md.imod().get());
                                dm.write();
                                mds.invalidate(imodPath);
                                Utils.LOGGER.info("Update completed");
                                found = true;
                            } catch (IOException e) {
                                throw new IOException("Update failed", e);
                            }
                        }
                    }
                    if (!found) Utils.LOGGER.error("Could not find any update");
                } catch (IOException e) {
                    throw new IOException("Could not delete " + mod, e);
                }
            }
        }
    }
}
